' /////////////////////////////////////////////////////////////////////////////
{ 
 Propeller C3 Test Suite prop_c3_unit_test_010.spin
 AUTHOR: Andre' LaMothe
 LAST MODIFIED: 9/13/10
 VERSION 1.0
 LICENSE: MIT
 

 COMMENTS:

 This test suite is to be ran on every C3 system off the manufacturing line.
 It exercises the system and tests everything out. The test consists of 3 major parts:

 1. Audio Visual Inspection Tests - for every unit, you plug it in, power it up, and
    then install this program. When the program runs, you will hear 3 tones on boot,
    you will see the NTSC display with white text on blue background as well as the VGA
    with a font dump and colored text. Finally, the BLUE power LED and RED status/VGA LED
    should be on. These tests you simply verify with your senses.

 2. Self Tests - for every unit, the software will test the SRAMs, FLASH, and SD card. Make
    sure a FAT16 formatted SD card is plugged into the C3. These tests will run and you will
    see a PASS/FAIL for each. The SD card is not written to, just accessed and tested, so
    you can use the same card for each unit and not worry about data being overwritten.

 3. Interactive Tests - these tests verify the PS/2 port is working, the IO ports (via a NES controller)
    as well as the A/D ports. For this test make sure your PS/2 keyboard is plugged in, the Parallax NES
    adapter is plugged in ONLY one row, flush with bottom of Port A (by SD card) and plug a NES controller
    into the upper port. Finally, plug in some kind of test rig into the A/D port that supplies a 0..5V
    voltage to channel 0,1 ( a pair of POTs powered from the port is great, I have supplied you with one).
    The tests are manual, so type some keys on the keyboard, spin the POTs, press some buttons on the NES
    controller, as you do, you will see the values written to the screen. Make sure they are correct.

4.  Power test, when you have completed the tests, pull the 9V supply, leaving the USB connected the C3
    should still be running without a glitch, then switch it off, pull all the cables, SHIP.

    That's it! -- 


 The following hardware is
 exercised:
  
 - 32K Bytes SRAMs Bank 0/1
 - 1MByte FLASH   
 - 2 Channel A/D Converter
 - IO Ports
 - PS/2 Port (Keyboard test)
 - VGA, NTSC, AUDIO
 - Serial UART
 - Power LED, Status/VGA LED
 - SD Card 
 - Voltage regulators
 - Game Pad with Parallax NES adapter part #32368, plug into P7-P0 port as follows:


                  CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

 Note: ONLY insert ONE row of the header into the C3, otherwise the other row on the
 adapter will short into the SPI, I2C, etc.

 REQUIRES: For all tests, you will need the following hardware. 

        (0) 9V DC power supply connected to power port.
        (1) USB port connected to PC for programming.
        (2) NTSC monitor/TV connected to A/V port.
        (3) VGA monitor connected to VGA port. 
        (4) PS/2 keyboard plugged in.
        (5) Parallax NES Gamepad Adapter and NES controller plugged into port nearest PS/2 connector (so A/D is easy to get to)
        (6) Micro SD card formatted FAT16, doesn't matter what is on it, the test will mount and access the card, nothing else.
        (7) A/D test rig, basically 2-pots, powered by +5 and GND with wipers going to channel 0/1 of AD
            - I have sent a sample rig for you to use to test the AD port, make sure to plug it in with BLACK / GND to the
            - bottom of the A/D port (GND).    
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal1 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  ' These settings are for 10 MHZ XTALS
  '_clkmode = xtal1 + pll8x          ' enable external clock and pll times 8
  '_xinfreq = 10_000_000             ' set frequency to 10 MHZ
  'CLOCKS_PER_MICROSECOND = 10*8       ' simple xin*pll / 1_000_000                     
  'CLOCKS_PER_MILLISECOND = 10_000*8   ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

   ' SPI (serial peripheral interface) interface pins, secondary names for SD driver functions
   spiDO  = 10
   spiCLK = 11
   spiDI  = 9

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

  ' SRAM function helper constants
  SPI_MODE_BYTE       = %00_00000_1
  SPI_MODE_PAGE       = %10_00000_1
  SPI_MODE_SEQUENTIAL = %01_00000_1


'//////////////////////////////////////////////////////////////////////////////
' GPU DEFINES /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' advanced GFX commands for GFX tile engine
  GPU_GFX_BASE_ID         = 192                  ' starting id for GFX commands to keep them away from normal command set
  GPU_GFX_NUM_COMMANDS    = 37                   ' number of GFX commands
   
  GPU_GFX_SUBFUNC_STATUS_R = (0+GPU_GFX_BASE_ID) ' Reads the status of the GPU, Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.                                                      
  GPU_GFX_SUBFUNC_STATUS_W = (1+GPU_GFX_BASE_ID) ' Writes status of the GPU, Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.

    ' sub-function constants that are executed when the GPU_GFX_SUBFUNC_STATUS_W command is issued 
    GPU_GFX_SUBFUNC_COPYMEM16 = 0 ' Copies numbytes from src -> dest in wordsize chunks
    GPU_GFX_SUBFUNC_FILLMEM16 = 1 ' Fills memory with data16, 2 bytes at a time

    GPU_GFX_SUBFUNC_COPYMEM8  = 2 ' Copies numbytes from src -> dest in byte size chunks
    GPU_GFX_SUBFUNC_FILLMEM8  = 3 ' Fills memory with low byte of data16, 1 bytes at a time 

  ' normal commands
     
  GPU_GFX_TILE_MAP_R = (2+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current tile map displayed.
  GPU_GFX_TILE_MAP_W = (3+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current tile map displayed.                                                                         
   
  GPU_GFX_DISPLAY_PTR_R = (4+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current display ptr into the tile map, low level print functions use this pointer.
  GPU_GFX_DISPLAY_PTR_W = (5+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current display ptr into the tile map, low level print functions use this pointer.                                                                
   
  GPU_GFX_TERM_PTR_R = (6+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current terminal ptr into the tile map, terminal level print functions use this pointer.
  GPU_GFX_TERM_PTR_W = (7+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current terminal ptr into the tile map, terminal level print functions use this pointer.                                                                
   
  GPU_GFX_BITMAP_R   = (8+GPU_GFX_BASE_ID) ' Reads 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap. 
  GPU_GFX_BITMAP_W   = (9+GPU_GFX_BASE_ID) ' Reads 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap.                                                                         
   
  GPU_GFX_PALETTE_R  = (10+GPU_GFX_BASE_ID) ' Reads 16-bit palette array ptr which points to the palettes in use for the tilemap.
  GPU_GFX_PALETTE_W  = (11+GPU_GFX_BASE_ID) ' Writes 16-bit palette array ptr which points to the palettes in use for the tilemap.                                                                         
   
  GPU_GFX_TOP_OVERSCAN_COL_R = (12+GPU_GFX_BASE_ID) ' Reads top overscan color drawn on the screen.
  GPU_GFX_TOP_OVERSCAN_COL_W = (13+GPU_GFX_BASE_ID) ' Writes top overscan color drawn on the screen.
   
  GPU_GFX_BOT_OVERSCAN_COL_R = (14+GPU_GFX_BASE_ID) ' Reads top overscan color drawn on the screen.
  GPU_GFX_BOT_OVERSCAN_COL_W = (15+GPU_GFX_BASE_ID) ' Writes top overscan color drawn on the screen.
   
  GPU_GFX_HSCROLL_FINE_R  = (16+GPU_GFX_BASE_ID)  ' Reads current fine horizontal scroll register (0..7) NOTE: (functionality NOT implemented yet) 
  GPU_GFX_HSCROLL_FINE_W  = (17+GPU_GFX_BASE_ID)  ' Writes current fine horizontal scroll register (0..7) NOTE: (functionality NOT implemented yet)
   
  GPU_GFX_VSCROLL_FINE_R  = (18+GPU_GFX_BASE_ID) ' Reads current fine vertical scroll register (0..7)
  GPU_GFX_VSCROLL_FINE_W  = (19+GPU_GFX_BASE_ID) ' Writes current fine vertical scroll register (0..7)
   
  GPU_GFX_SCREEN_WIDTH_R  = (20+GPU_GFX_BASE_ID)  ' Reads screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc.
  GPU_GFX_SCREEN_WIDTH_W  = (21+GPU_GFX_BASE_ID)  ' Writes screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc.
   
  GPU_GFX_SRC_ADDR_R      = (22+GPU_GFX_BASE_ID) ' Reads 16-bit source address for GPU operations.
  GPU_GFX_SRC_ADDR_W      = (23+GPU_GFX_BASE_ID) ' Writes 16-bit source address for GPU operations.
                       
  GPU_GFX_DEST_ADDR_R     = (24+GPU_GFX_BASE_ID) ' Reads 16-bit destination address for GPU operations.
  GPU_GFX_DEST_ADDR_W     = (25+GPU_GFX_BASE_ID) ' Writes 16-bit destination address for GPU operations.
   
  GPU_GFX_NUM_BYTES_R     = (26+GPU_GFX_BASE_ID) ' Reads 16-bit number representing the number of bytes for a GPU operation Sub-Function.
  GPU_GFX_NUM_BYTES_W     = (27+GPU_GFX_BASE_ID) ' Writes 16-bit number representing the number of bytes for a GPU operation Sub-Function.
   
  GPU_GFX_DATA_R          = (28+GPU_GFX_BASE_ID) ' Reads 16-bit data word uses for GPU operations or memory access operations.
  GPU_GFX_DATA_W          = (29+GPU_GFX_BASE_ID) ' Writes 16-bit data word uses for GPU operations or memory access operations.
   
  GPU_GFX_RAM_PORT8_R     = (30+GPU_GFX_BASE_ID) ' Reads 8-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                           ' After the operation, the src_addr_ptr is incremented as per the GPU configuration.
   
  GPU_GFX_RAM_PORT8_W     = (31+GPU_GFX_BASE_ID) ' Writes 8-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                           ' After the operation, the src_add_rptr are incremented as per the GPU configuration.
   
  GPU_GFX_RAM_PORT16_R    = (32+GPU_GFX_BASE_ID) ' Reads 16-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                           ' After the operation, the src_addr_ptr is incremented as per the GPU configuration.
   
  GPU_GFX_RAM_PORT16_W    = (33+GPU_GFX_BASE_ID) ' Writes 16-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                           ' After the operation, the src_add_rptr are incremented as per the GPU configuration.
   
  GPU_GFX_RASTER_LINE_R   = (34+GPU_GFX_BASE_ID) ' Reads the current raster line being drawn from 0..191, or whatever the GPU's line resolution is set to.                             
   
  ' gpu configuration registers
  GPU_GFX_SET_AUTO_INC_R    = (35+GPU_GFX_BASE_ID) ' Reads current memory auto increment setting for read/write operations lower 4-bits (0..15), default 0
  GPU_GFX_SET_AUTO_INC_W    = (36+GPU_GFX_BASE_ID) ' Writes the current memory auto increment setting for read/write operations lower 4-bits (0..15), default 0

  ' ASCII codes for ease of parser development
  ASCII_A      = 65
  ASCII_B      = 66
  ASCII_C      = 67
  ASCII_D      = 68
  ASCII_E      = 69
  ASCII_F      = 70
  ASCII_G      = 71
  ASCII_H      = 72
  ASCII_O      = 79  
  ASCII_P      = 80
  ASCII_Z      = 90
  ASCII_0      = 48
  ASCII_9      = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $C8 ' backspace
  ASCII_DEL    = $C9 ' delete
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $CB ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' null pointer, null character
  NULL         = 0  

'//////////////////////////////////////////////////////////////////////////////
' SD CARD DEFINES /////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

   ' SD card (secure digital) constants (you can reduce these to speed things up and see if the SD card still works)
   SD_RESPONSE_LIMIT        = 256     ' maximum number of iterations that SD card will be querried for response after a command sent
   SD_POWERUP_LIMIT         = 256     ' maximum number of iterations to initially clock SD card to power up
   SD_RETRIEVE_SECTOR_LIMIT = 4096    ' maximum number of iterations for SD to retrieve a requested sector/block
   SD_WRITE_SECTOR_LIMIT    = 128     ' maximum number of iterations for SD to write a requested sector/block

   ' SD commands (based on SD spec names)
   SD_GO_IDLE_STATE     = $40         ' resets the SD card, places it into SPI mode, idle                              
   SD_EXIT_IDLE_STATE   = $41         ' takes SD out of idle state ready for commands

   SD_APP_CMD           = ($40 + $37) ' makes sure card is in SD mode
   SD_SEND_OP_CMD       = ($40 + $29) ' makes sure card is in SD mode
   SD_READ_SINGLE_BLOCK = ($40 + $11) ' reads a single block (usually a 512 byte sector) at byte address sent in command address
   SD_WRITE_BLOCK       = ($40 + $18) ' write a single block (usually a 512 byte sector) at byte address sent in command address   

   SECTOR_SIZE          = 512         ' should always be 512 for SD cards                  
   MAX_SECTORS_BUFFER   = 1           ' number of sectors buffer can hold, make larger if needed 

   ' fat constants for various functions
   FILE_SEEK_ABS        = 0           ' interpret seek request with absolute position                                         
   FILE_SEEK_REL        = 1           ' interpret seek request as relative position added to current position
   FILE_SEEK_END        = 2           ' seek to end of file   (file_size - 1)
   FILE_SEEK_START      = 3           ' seek to start of file (0)


'//////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY DEFINES ////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

' defines for the proper FLASH installed on C3 4MB, 8MB, 16MB
' 4Mb (512K Bytes)
AT26F004        = 4  ' numeric codes for flash version
AT25DF041A      = 4

' 8Mb (1M Bytes)
AT26DF081A      = 8
AT25DF081A      = 8

' 16 Mb (2M Bytes)
AT25DF161       = 16

' set type of FLASH ROM currently installed in C3 (8MBit) for conditional compilation of code -- not supported in SPIN, so manually done below :( 
FLASH_ROM             = AT26DF081A 
NUM_FLASH_ROM_SECTORS = 19 ' 11 for 4MB, 19 for 8MB, must be set manually due to lack of conditional compilation


' Atmel 2F004/8xxx series basic commands
WRITE_ENABLE     = $06
WRITE_DISABLE    = $04
READ_STATUS      = $05
WRITE_STATUS     = $01
READ_DATA        = $03

SEQ_PROGRAM      = $AF  
BYTE_PROGRAM     = $02 

BLOCK_ERASE4     = $20
BLOCK_ERASE32    = $52
BLOCK_ERASE64    = $D8

SECTOR_ERASE     = $20    

CHIP_ERASE       = $60 
JEDEC_ID         = $9F

PROTECT_SECTOR   = $36 
UNPROTECT_SECTOR = $39 

' custom commands we make up, not part of chip spec, just used here, start at code 0xF0
UNPROTECT_CHIP   = $F0

'//////////////////////////////////////////////////////////////////////////////
' NES GAMEPAD DEFINES /////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

{
     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

}

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0

  NES_LATCH_DELAY  = $40

  ' button ids/bit masks
  ' NES bit encodings general for state bits
  NES_RIGHT  = %00000001
  NES_LEFT   = %00000010
  NES_DOWN   = %00000100
  NES_UP     = %00001000
  NES_START  = %00010000
  NES_SELECT = %00100000
  NES_B      = %01000000
  NES_A      = %10000000

  ' NES bit encodings for NES gamepad 0
  NES0_RIGHT  = %00000000_00000001
  NES0_LEFT   = %00000000_00000010
  NES0_DOWN   = %00000000_00000100
  NES0_UP     = %00000000_00001000
  NES0_START  = %00000000_00010000
  NES0_SELECT = %00000000_00100000
  NES0_B      = %00000000_01000000
  NES0_A      = %00000000_10000000

  ' NES bit encodings for NES gamepad 1
  NES1_RIGHT  = %00000001_00000000
  NES1_LEFT   = %00000010_00000000
  NES1_DOWN   = %00000100_00000000
  NES1_UP     = %00001000_00000000
  NES1_START  = %00010000_00000000
  NES1_SELECT = %00100000_00000000
  NES1_B      = %01000000_00000000
  NES1_A      = %10000000_00000000


   ' debugging constants for function terminal debug outputs for SD 
   DEBUG_OFF            = 2           ' no output at all
   DEBUG_VERBOSE        = 1           ' verbose mode for functions debug output
   DEBUG_BRIEF          = 0           ' brief mode for functions debug output

'//////////////////////////////////////////////////////////////////////////////
' VARS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////        
  
VAR

  long g_data, g_data2, g_status    
  long g_data32, g_data16, g_spi_result
  long g_key, g_temp1, g_temp2, g_temp3, g_temp4, g_index, g_length
  long g_tempx, g_tempy 
  byte g_sbuffer[512]

  ' sd card stuff
  long debug_level              ' debug level for SD driver
  long debug_wait_res_num_iter  ' the number of times the sd send command function internally waits for a SPI response
  
DAT

'  conditional compilation for FLASH memory variants not supported, so just comment out either declaration

' #if (FLASH_ROM == AT26F004)

'// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
'// 11 sectors [64K, 64K, 64K, 64K, 64K, 64K, 64K, 32K, 8K, 8k, 16K ]
'//            |<------- (7) 64K sectors ------->|  1      2     1 

{
    ' 11 sectors in 4MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 32K sector 7
                     long $78000   ' 8K sector 8  
                     long $7A000   ' 8K sector 9
                     long $7D000   ' 8K sector 10
}

' #if (FLASH_ROM == AT26DF081A)

' we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
' 19 sectors [64K, 64K, 64K, ..., 64K, 64K, 64K, 64K, 16K, 8K, 8k, 32K ]
'            |<----------- (15) 64K sectors -------->| 1     2     1 

    ' 19 sectors in 8MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 64K sector 7 
                     long $80000   ' 64K sector 8
                     long $90000   ' 64K sector 9
                     long $A0000   ' 64K sector 10
                     long $B0000   ' 64K sector 11
                     long $C0000   ' 64K sector 12
                     long $D0000   ' 64K sector 13
                     long $E0000   ' 64K sector 14
                          
                     long $F0000   ' 16K sector 15
                     long $F4000   '  8K sector 16
                     long $F6000   '  8K sector 17
                          
                     long $F8000    ' 32K sector 18




'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

' current drivers used in this version of SPI interface 
'term_ntsc  : "TV_Text_Half_Height_011.spin"        ' instantiate NTSC terminal driver

term_vga   : "VGA_Text_010.spin"                   ' instantiate VGA terminal driver
kbd        : "keyboard_010.spin"                   ' instantiate keyboard driver
snd        : "NS_sound_drv_052_11khz_16bit.spin"   ' instantiate sound driver         
gfx_ntsc   : "C3_GFX_DRV_010.SPIN"                 ' instantiate new NTSC tile driver


'//////////////////////////////////////////////////////////////////////////////
' PUBS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Main : status

  ' let the system initialize 
  Delay_US( 1*1_000_000 )

  ' initialize VGA terminal
  term_vga.start(%10111)   

  ' initialize the graphics tile engine
  gfx_ntsc.start

  ' start sound engine up
  snd.start( AUDIO_MONO )

  ' delay a moment
  Delay_US(1*10_000)

  ' make a little noise
  snd.PlaySoundFM(0, snd#SHAPE_SQUARE, snd#NOTE_C4, CONSTANT( Round(Float(snd#SAMPLE_RATE) * 0.25)), 200, $3579_ADEF )
  repeat 50_000
  snd.PlaySoundFM(1, snd#SHAPE_SQUARE, snd#NOTE_C5, CONSTANT( Round(Float(snd#SAMPLE_RATE) * 0.25)), 200, $3579_ADEF )
  repeat 50_000
  snd.PlaySoundFM(2, snd#SHAPE_SQUARE, snd#NOTE_C6, CONSTANT( Round(Float(snd#SAMPLE_RATE) * 0.25)), 200, $3579_ADEF )

  'start keyboard on chameleon pins, keyboard is always started even if its not plugged in
  kbd.start( PS2_DATA, PS2_CLK)


  ' test out VGA hardware, display some characters, test buffer chip
  term_vga.Out( $00 )
  term_vga.newline  
  term_vga.pstring( string ("C3 Manufacturing Q/A Suite V1.0") )
  term_vga.newline
  term_vga.newline

  term_vga.pstring( string ("Font Dump:") )
  term_vga.newline
  term_vga.newline


  term_vga.pstring( string ("abcdefghijklmnopqrstuvwxyz") )
  term_vga.newline

  term_vga.pstring( string ("ABCDEFGHIJKLMNOPQRSTUVWXYZ") )
  term_vga.newline

  term_vga.pstring( string ("~!@#$%^&*()_+01234567890[]{}") )
  term_vga.newline

  term_vga.pstring( string (";':,.<>/?") )
  term_vga.newline

  term_vga.Out( $0C )
  term_vga.Out( $01 )
  term_vga.pstring( string ("Yellow on yellow.") )
  term_vga.newline

  term_vga.Out( $0C )
  term_vga.Out( $02 )
  term_vga.pstring( string ("Magneta on black.") )
  term_vga.newline

  term_vga.Out( $0C )
  term_vga.Out( $03 )
  term_vga.pstring( string ("Gray on white.") )
  term_vga.newline

  term_vga.Out( $0C )
  term_vga.Out( $04 )
  term_vga.pstring( string ("Cyan on green.") )
  term_vga.newline

  
 ' /////////////////////////////////////////////////////////////////////////////
  ' INITIALIZE I/O PINS /////////////////////////////////////////////////////////
  ' /////////////////////////////////////////////////////////////////////////////

  ' blink the status LED 3 times to show board is "alive", then enable VGA (leave LOW)
  DIRA[STATUS_LED_BUS_MUX] := 1 ' set to output
  OUTA[STATUS_LED_BUS_MUX] := 0

  repeat 6
    OUTA[STATUS_LED_BUS_MUX] := !OUTA[STATUS_LED_BUS_MUX]
    repeat 25_000 

  ' turn on VGA buffer
  OUTA[STATUS_LED_BUS_MUX] := 0

  ' initialize SPI bus system
  SPI_Init

  ' initialize gamepad IO
  NES_Init_Gamepad

  ' init A/D converter
  AD_Init

 ' initialize SRAM
  SRAM_Init( SPI_MODE_SEQUENTIAL )

  ' open the flash memory up (unprotect it)
  Flash_Open

  ' set debug level for SD functions
  debug_level := DEBUG_OFF
  
  ' begin test loop

  ' clear screen
  gfx_ntsc.Out_Term( $00 )

  repeat
    ' draw menu

    gfx_ntsc.String_Term( string ("C3 Manufacturing Q/A Suite V1.0") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term
    
    gfx_ntsc.String_Term( string ("[AUDIO/VISUAL INSPECTION TESTS]") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string (" LEDs: Power(Blue),VGA(Red)?") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string (" UART LEDs: TX(Green),RX(Red)?") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string (" NTSC: White text on blue?") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string (" VGA: Character dump+colors?") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string (" AUDIO: 3 tones on boot?") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string ("[SELF TESTS...]") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string (" FLASH 1MB: ") )
    ' the FLASH is tested from manufacturer, thus we are not going to do an exaustive test,
    ' rather a simplified functional test. We are going to erase the FLASH memory, verify
    ' the first 256 bytes are indeed erased, then write a pattern of 0..255 to the first sector
    ' and verify it.

    g_tempx := gfx_ntsc.GetX
    g_tempy := gfx_ntsc.GetY
    gfx_ntsc.String_Term( string ("Erasing...") )
    Flash_Erase_Chip(1)
 
    gfx_ntsc.GotoXY( g_tempx, g_tempy)
    gfx_ntsc.String_Term( string ("Writing...") )

    repeat g_index from 0 to 511
      g_sbuffer[ g_index ] := (g_index*g_index) // 256
      
    ' write buffer to flash
    Flash_Write( 0, @g_sbuffer, 512 )

    ' clear buffer out
    bytefill ( @g_sbuffer, $00, 512) 
    
    gfx_ntsc.GotoXY( g_tempx, g_tempy)
    gfx_ntsc.String_Term( string ("Reading...") )

    ' read flash into buffer
    Flash_Read( 0, @g_sbuffer, 512 )

    ' reset error flag
    g_temp1 := 0

    gfx_ntsc.GotoXY( g_tempx, g_tempy)
    gfx_ntsc.String_Term( string ("Verifying...") )

    ' check each byte
    repeat g_index from 0 to 511
      if (g_sbuffer[ g_index ] <> (g_index*g_index) // 256)
        ' error!!!!
        g_temp1 := 1
        quit ' exit repeat loop

    ' position cursor
    gfx_ntsc.GotoXY( g_tempx, g_tempy)
      
    ' test for error flag?
    if (g_temp1 == 0)
      gfx_ntsc.String_Term( string ("PASSED.     ") )
    else
      gfx_ntsc.String_Term( string ("FAILED!!!   ") )

    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string (" SRAM 0|1: ") )
    ' the SRAMs are tested from manufacturer, so we are not going to do an exaustive test
    ' rather a basic functional verification test, we are going to write a pattern to each 
    ' sram of $FF, $00, $AA, $55 and verify it reads back out


' ///////////////////////////////////////////////////////

    g_tempx := gfx_ntsc.GetX
    g_tempy := gfx_ntsc.GetY
    gfx_ntsc.String_Term( string ("Bank0") )

    ' error flag
    g_temp3 := 0
    g_index := 0
        
    repeat while ((g_temp3 == 0) and (g_index < 4) )
      gfx_ntsc.Out_Term( "." )
      ' retrieve test byte
      g_temp1 := sram_test_vector[g_index]

      ' write data 
      SRAM_Fill( 0, 0, 256, g_temp1)

      ' read data and test
      SRAM_Read(0, 0, 256, @g_sbuffer)    
      
      ' check each byte
      repeat g_temp2 from 0 to 255
        if (g_sbuffer[ g_temp2 ] <> g_temp1)
          ' error!!!!
          g_temp3 := 1
          quit ' exit repeat loop
          
      ' update test index
      g_index++

    ' position cursor
    gfx_ntsc.GotoXY( g_tempx, g_tempy)
      
    ' test for error flag?
    if (g_temp3 == 0)
      gfx_ntsc.String_Term( string ("PASSED.  ") )
    else
      gfx_ntsc.String_Term( string ("FAILED!!!") )


' ///////////////////////////////////////////////////////

    g_tempx := gfx_ntsc.GetX
    g_tempy := gfx_ntsc.GetY
    gfx_ntsc.String_Term( string ("Bank1") )

    ' error flag
    g_temp3 := 0
    g_index := 0
    
    repeat while ((g_temp3 == 0) and (g_index < 4) )
      gfx_ntsc.Out_Term( "." )
      ' retrieve test byte
      g_temp1 := sram_test_vector[g_index]

      ' write data 
      SRAM_Fill( 1, 0, 256, g_temp1)

      ' read data and test
      SRAM_Read(1, 0, 256, @g_sbuffer)    
      
      ' check each byte
      repeat g_temp2 from 0 to 255
        if (g_sbuffer[ g_temp2 ] <> g_temp1)
          ' error!!!!
          g_temp3 := 1
          quit ' exit repeat loop
          
      ' update test index
      g_index++

    ' position cursor
    gfx_ntsc.GotoXY( g_tempx, g_tempy)
      
    ' test for error flag?
    if (g_temp3 == 0)
      gfx_ntsc.String_Term( string ("PASSED.  ") )
    else
      gfx_ntsc.String_Term( string ("FAILED!!!") )


    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string (" SD Card: ") )
    ' the SD card interface is nothing more than a few wires to the SD mechanical header, so
    ' testing the card read/write of sectors isn't useful, since you would be testing the CARD
    ' plugged into the header, rather, we simply want to confirm all the header solder points
    ' are down and that the card is READABLE -- this IS useful, since the SD card mechanical is
    ' hard to solder. Thus, we will simply run a "mount" test on the SD card.

    g_temp1 := SD_Mount

    ' test for error flag?
    if (g_temp1 == $00)
      gfx_ntsc.String_Term( string ("PASSED - Mounted.  ") )
    else
      gfx_ntsc.String_Term( string ("FAILED!!!") )


    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string ("[INTERACTIVE TESTS]") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term


    ' save current cursor position 
    g_tempx := gfx_ntsc.GetX
    g_tempy := gfx_ntsc.GetY

    ' interactive tests are in real time repeat forever...
    repeat 

      gfx_ntsc.GotoXY( g_tempx, g_tempy)

      gfx_ntsc.String_Term( string (" Gamepad(s) Button State:") )
      gfx_ntsc.Newline_Term
      
      ' read the gamepads
      g_temp1 := NES_Read_Gamepad

      ' print results in binary
      gfx_ntsc.String_Term( string ("  GP0[") )
      gfx_ntsc.Bin_Term( (g_temp1 ) & $FF, 8 )
      gfx_ntsc.String_Term( string ("],") )

      gfx_ntsc.String_Term( string ("GP1[") )      
      gfx_ntsc.Bin_Term( (g_temp1 >> 8) & $FF, 8 )
      gfx_ntsc.String_Term( string ("] ") )
      gfx_ntsc.Newline_Term
      gfx_ntsc.Newline_Term
       
      gfx_ntsc.String_Term( string (" Keyboard Key: ") )

      if (kbd.gotkey == TRUE)
        ' print char to screen
        gfx_ntsc.Out_Term( "`" )
        gfx_ntsc.Out_Term( kbd.key )   
        gfx_ntsc.Out_Term( "'" )
      
      gfx_ntsc.Newline_Term
      gfx_ntsc.Newline_Term
             
      gfx_ntsc.String_Term( string (" A/D Data: ") )
      ' read channel 0
      gfx_ntsc.String_Term( string ("Ch0=") )
      gfx_ntsc.Dec_Term( AD_Read( 0 ) )
      gfx_ntsc.Out_Term( "," )  
      gfx_ntsc.String_Term( string ("Ch1=") )
      gfx_ntsc.Dec_Term( AD_Read( 1 ) )
      gfx_ntsc.String_Term( string ("  ") )       


    ' get string from user
    Get_String(@g_sbuffer, 4 )

    ' convert to integer
    g_temp1 := atoi2(@g_sbuffer, 4)
 
' end Main

' /////////////////////////////////////////////////////////////////////////////

PUB Get_String(string_ptr, max_length) | _length
{{
This simple function is a single line editor that allows user to enter keys from the keyboard
and then echos them to the screen, when the user hits <ENTER> | <RETURN> the function
exits and returns the string. The function has simple editing and allows <BACKSPACE> to
delete the last character, that's it!

The function outputs to the NTSC terminal

PARMS:    

RETURNS: length of string, 0 if NULL.

}}

  ' current length of string buffer
  _length := 0  

  ' draw cursor
  
  repeat 

    ' draw cursor
    gfx_ntsc.Out_Term( "_" )
    gfx_ntsc.Out_Term( $08 )
  
    ' wait for keypress 
    repeat while (kbd.gotkey == FALSE)

    ' user entered a key process it

    ' get key from buffer
    g_key := kbd.key
     
    case g_key
       ASCII_LF, ASCII_CR: ' return    
 
        ' null terminate string and return
        byte [string_ptr][_length] := ASCII_NULL
     
        return( _length )


       ASCII_BS, ASCII_DEL, ASCII_LEFT: ' backspace (edit)

         if (_length > 0)
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )          
           gfx_ntsc.Out_Term( $08 )
           
           ' echo character
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )
         
           ' decrement length
           _length--
 
       other:    ' all other cases
         ' insert character into string 
         byte [string_ptr][_length] := g_key

         ' update length
         if (_length < max_length )
           _length++
         else
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( $08 )          

         ' echo character
         gfx_ntsc.Out_Term( g_key )
     
' // end Get_String


' /////////////////////////////////////////////////////////////////////////////

PUB StrCpy( dest_str_ptr, source_str_ptr ) | _index
{{
mimics C strcpy, copies source -> destination including NULL
}}

' test if there is storage
if (dest_str_ptr == NULL)
  return (NULL)

_index := 0

repeat while (byte [ source_str_ptr ][_index] <> NULL)
  ' copy next byte
   byte [ dest_str_ptr ][_index] := byte [ source_str_ptr ][_index]    
   _index++

' null terminate
byte [ dest_str_ptr ][_index] := NULL

' return number of bytes copied
return ( _index )

' // end StrCpy

' /////////////////////////////////////////////////////////////////////////////

PUB StrUpper(string_ptr)
{{
upcases a string
}}

  if (string_ptr <> NULL)
    repeat while (byte[ string_ptr ] <> NULL)
      byte[ string_ptr ] :=  ToUpper( byte[ string_ptr ] )
      string_ptr++  

  ' return string
  return (string_ptr)

' // end StrUpper

' /////////////////////////////////////////////////////////////////////////////

PUB ToUpper(ch)
{{
returns the uppercase of the sent character
}}

if (ch => $61 and ch =< $7A)
  return(ch-32)
else
  return(ch)

' // end ToUpper

' /////////////////////////////////////////////////////////////////////////////

PUB IsInSet(ch, set_string)
{{
tests if sent character is in string
}}

repeat while (byte[set_string][0] <> NULL)
  if (ch == byte[set_string++][0])
    'serial.txstring(string ("found!") )
    return( ch )

'serial.txstring(string ("not found!") )
' not found
return( -1 )    

' // end IsInSet

' /////////////////////////////////////////////////////////////////////////////

PUB IsSpace(ch)
{{
tests if sent character is white space, cr, lf, space, tab
}}

if ( (ch == ASCII_SPACE) or (ch == ASCII_LF) or (ch == ASCII_CR) or (ch == ASCII_TAB))
  return (ch)
else
  return(-1)  

' // end IsSpace

' /////////////////////////////////////////////////////////////////////////////

PUB IsNull(ch)
{{
tests if sent character is null
}}

if ( (ch == NULL))
  return (1)
else
  return(-1)  

' // end IsNull

' /////////////////////////////////////////////////////////////////////////////

PUB IsDigit(ch)
{{
tests if sent character is a number 0..9, returns integer 0..9
}}

if (ch => ASCII_0 and ch =< ASCII_9)
  return (ch-ASCII_0)
else
  return(-1)  

' // end IsDigit

' /////////////////////////////////////////////////////////////////////////////

PUB IsAlpha(ch)
{{
tests if sent character is a number a...zA...Z
}}

ch := ToUpper(ch)

if ( (ch => ASCII_A) and (ch =< ASCII_Z))
  return (ch)
else
  return(-1)  

' end IsAlpha

' /////////////////////////////////////////////////////////////////////////////

PUB IsPunc(ch)
{{
tests if sent character is a punctuation symbol !@#$%^&*()--+={}[]|\;:'",<.>/?
}}

ch := ToUpper(ch)

if (  ((ch => 33) and (ch =< 47)) or  ((ch => 58) and (ch =< 64)) or ((ch => 91) and (ch =< 96)) or ((ch =>123) and (ch =< 126))  ) 
  return (ch)
else
  return(-1)  

' // end IsPunc
                                                                                
' /////////////////////////////////////////////////////////////////////////////

PUB HexToDec(n)
{{
converts hex digit to decimal
}} 
  if ( (n => "0") and (n =< "9") )
    return (n - ASCII_0)
  elseif ( (n => "A") and (n =< "F") )
    return (n - "A" + 10)
  else
    return ( 0 )  

' // end HexToDec

' /////////////////////////////////////////////////////////////////////////////

PUB itoa(value, string_ptr) | i
{{
 converts value into a string and stores in string_ptr with NULL termination
 }}

  if value < 0
    -value
    byte [ string_ptr++ ]  := "-"

  i := 1_000_000_000

  repeat 10
    if value => i
      byte [ string_ptr++ ]  := value / i + "0"
      value //= i
      result~~
    elseif result or i == 1
      byte [ string_ptr++ ]  := "0"
    i /= 10

  ' null terminate
  byte [ string_ptr ] := NULL    

' // end itoa

' /////////////////////////////////////////////////////////////////////////////

PUB atoi2(string_ptr, length) | _index, _sum, _ch, _sign
{{
this function tries to convert the string to a number, supports binary %, hex $, decimal (default)
string_ptr can be to a null terminated string, or the length can be overridden by sending the length shorter
than the null terminator, ignores ws
Eg. %001010110, $EF, 25
}}

' initialize vars
_index := 0
_sum   := 0
_sign  := 1
 
' consume white space
repeat while (IsSpace( byte[ string_ptr ][_index] ) <> -1)
  _index++
 
' is there a +/- sign?
if (byte [string_ptr][_index] == "+")
  ' consume it
  _index++
elseif (byte [string_ptr][_index] == "-")
  ' consume it
  _index++
  _sign := -1    
     
' try to determine number base
if (byte [string_ptr][_index] == ASCII_HEX)
  _index++
  repeat while ( ( IsDigit(_ch := byte [string_ptr][_index]) <> -1) or ( IsAlpha(_ch := byte [string_ptr][_index]) <> -1)  )
    _index++
    _sum := (_sum << 4) + HexToDec( ToUpper(_ch) )
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if hex number    
elseif (byte [string_ptr][_index] == ASCII_BIN)
  repeat while ( IsDigit(_ch := byte [string_ptr][++_index]) <> -1)
    _sum := (_sum << 1) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if binary number  
else
  ' must be in default base 10, assume that
  repeat while ( IsDigit(_ch := byte [string_ptr][_index++]) <> -1)
    _sum := (_sum * 10) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
 
' else, have no idea of number format!
return(0)
 
' // end atoi2

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US

' /////////////////////////////////////////////////////////////////////////////
' SD CARD SPI MODE MINIMAL ROUTINES
' /////////////////////////////////////////////////////////////////////////////

PUB SD_Mount | index, index2, sd_response, cmd_num_attempts 

' DESCRIPTION: This function mounts the SD card which is a fairly complex process consisting of a number of steps.
' 1. Power up.
' 2. SD card must be placed into "idle state"
' 3. Then taken out of idle state.
' 4. Next, the card is tested to make sure its an SD and not MMC.
' 5. Finally, if all passes then the card is ready to go and "mounted".
'
' INPUTS: none
 '
' OUTPUTS: returns $00 for success, $FF for failure 
' -----------------------------------------------------------------------------

  if (debug_level == DEBUG_BRIEF)
    gfx_ntsc.String_Term(STRING("Mounting SD card..."))
    gfx_ntsc.Out_Term($0D)

   ' de-assert chip select
  'OUTA [ spiCS ]  := 1
  SPI_Select_Channel( 0 )

  
  ' now clock the SD SPI interface a few times and let it initialize (send dummy data)
  repeat index from 0 to SD_POWERUP_LIMIT
    SPI_Send_Recv_Byte( $FF )

  ' assert chip select
  'OUTA [ spiCS ]  := 0
  SPI_Select_Channel( 5 )

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    gfx_ntsc.String_Term(STRING("Placing SD card into idle state..."))
    gfx_ntsc.Out_Term($0D)
  
  sd_response := $FF
  cmd_num_attempts := 0
  repeat while (sd_response <> $01)

    ' send SD command GO_IDLE_STATE = $40, which resets the SD card
    sd_response := SD_Send_Command( SD_GO_IDLE_STATE, 0 )  

    ' test for verbose mode
    if (debug_level == DEBUG_VERBOSE)
      ' print response
      gfx_ntsc.String_Term(STRING("Response = "))
      gfx_ntsc.Hex_Term(sd_response, 2)
      gfx_ntsc.Out_Term($0D)
       
      gfx_ntsc.String_Term(STRING("Number of iterations = "))
      gfx_ntsc.Dec_Term(debug_wait_res_num_iter)
      gfx_ntsc.Out_Term($0D)
       
      gfx_ntsc.String_Term(STRING("Number of attempts = "))
      gfx_ntsc.Dec_Term(cmd_num_attempts)
      gfx_ntsc.Out_Term($0D)
       
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
      ' test for verbose mode
      if (debug_level == DEBUG_VERBOSE)
        gfx_ntsc.String_Term(STRING("SD card not responding to idle state command."))
        gfx_ntsc.Out_Term($0D)

      if (debug_level == DEBUG_BRIEF)
        gfx_ntsc.String_Term(STRING("Couldn't mount SD card!"))
        gfx_ntsc.Out_Term($0D)

      return ($FF)

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    gfx_ntsc.String_Term(STRING("Bringing SD card out of idle state..."))
    gfx_ntsc.Out_Term($0D)
     
  sd_response := $FF
  cmd_num_attempts := 0  
  repeat while (sd_response <> $00)

    ' send SD command EXIT_IDLE_STATE = $41, which resets the SD card
    sd_response := SD_Send_Command( SD_EXIT_IDLE_STATE, 0 )  

    ' test for verbose mode
    if (debug_level == DEBUG_VERBOSE)
      ' print response
      gfx_ntsc.String_Term(STRING("Response = "))
      gfx_ntsc.Hex_Term(sd_response, 2)
      gfx_ntsc.Out_Term($0D)
       
      gfx_ntsc.String_Term(STRING("Number of iterations = "))
      gfx_ntsc.Dec_Term(debug_wait_res_num_iter)
      gfx_ntsc.Out_Term($0D)
       
      gfx_ntsc.String_Term(STRING("Number of attempts = "))
      gfx_ntsc.Dec_Term(cmd_num_attempts)
      gfx_ntsc.Out_Term($0D)
       
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
      ' test for verbose mode
      if (debug_level == DEBUG_VERBOSE)
        gfx_ntsc.String_Term(STRING("SD card not responding to exit idle state command."))
        gfx_ntsc.Out_Term($0D)

      if (debug_level == DEBUG_BRIEF)
        gfx_ntsc.String_Term(STRING("Couldn't mount SD card!"))
        gfx_ntsc.Out_Term($0D)
    
      return ($FF)

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    gfx_ntsc.String_Term(STRING("Verify SD card mode (not a MMC)..."))
    gfx_ntsc.Out_Term($0D)
     
  ' send SD command APP_CMD = $40 + $37
  sd_response := SD_Send_Command( SD_APP_CMD, 0 )

  if ( sd_response <> $00 )
    if (debug_level == DEBUG_BRIEF)
      gfx_ntsc.String_Term(STRING("Couldn't mount SD card!"))
      gfx_ntsc.Out_Term($0D)
      
    return($FF)

  ' send SD command SEND_OP_CMD = $40 + $29
  sd_response := SD_Send_Command( SD_SEND_OP_CMD, 0 )

  if ( sd_response <> $00 )
    if (debug_level == DEBUG_BRIEF)
      gfx_ntsc.String_Term(STRING("Couldn't mount SD card!"))
      gfx_ntsc.Out_Term($0D)

    return($FF)

  if (debug_level == DEBUG_BRIEF)
    gfx_ntsc.String_Term(STRING("SD card successfully mounted."))
    gfx_ntsc.Out_Term($0D)

  ' return success
  return($00)

' end SD_Mount

' /////////////////////////////////////////////////////////////////////////////

PUB SD_Send_Command( command8, address32 ) | response, index

' DESCRIPTION: This function is the main workhorse of the SD interface and is used to send commands
' to the SD card via the SPI interface. All SD commands consist of 6 bytes in the following format:
'
' Command id     (1-byte) This you get from the SD specification.
' 32-bit address (4-bytes) In big endian form.
' CRC result     (1-byte) Error checking byte.
' 
' The address is in big endian form, that is high byte to low byte (opposite of PC and propeller)
' The cycle redundancy check byte is sent in CRC7 or CRC16 format for all transactions
' except status tokens. CRC is of course a method of error detection, but not
' correction. If a CRC fails to match then the data must be sent or requested again.
' the algorithms for CRC7 and CRC16 are well documented and are basically division algorithms
' where the CRCs are the remainders. In our case, we are always going to send $95 which
' is the CRC for the first command that will be sent that matters, after that CRCs are
' ignored in SPI mode, later add code to compute the CRC for kicks if you like.
' Also, note the use of SD_RESPONSE_LIMIT, this allows the command to be sent over and over until there is a response.
' but, if the attempts exceed this constant, there is something wrong with the SD card.
'
' INPUTS: command8 - 8-bit command from SD card specification refer to document "SD Card Specification" from SD Association.
'         address32 - 32-bit address in normal little endian format, function converts to big endian before transport
'
' OUTPUTS: returns status of command to SD card, typically $FF means error, but for each command response can be different
' -----------------------------------------------------------------------------

  ' first send command
  SPI_Send_Recv_Byte(command8)  

  ' now send 4-bytes of address in MS-Byte to LS-byte format
  SPI_Send_Recv_Byte( (address32 >> 24) & $FF)  
  SPI_Send_Recv_Byte( (address32 >> 16) & $FF)
  SPI_Send_Recv_Byte( (address32 >> 8)  & $FF)
  SPI_Send_Recv_Byte( (address32 >> 0)  & $FF) 

  ' finally the CRC
  SPI_Send_Recv_Byte( $95 )            

  ' now wait for response
  response := $FF ' default response is $FF which means error or time out

  repeat index from 0 to SD_RESPONSE_LIMIT
    ' read response by sending dummy data
    response := SPI_Send_Recv_Byte( $FF )
  
    ' DEBUG CODE - send out number of iterations the response took to get back
    debug_wait_res_num_iter := index

    ' test if we have a non-$FF response
    if (response <> $FF)
      ' send dummy data to allow SD card to finish operation, needed after all commands
      SPI_Send_Recv_Byte ( $FF )
      ' return results
      return ( response )
      
  ' if we got here then something bad happened!
  return ( $FF )

' end SD_Send_Command(command8, address32)

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Send_Recv_Byte(spi_data8) | spi_result8, index

' DESCRIPTION: this functions sends spi_data8 and receives spi_result8, remember all spi operations are circular, so
' when a byte is sent a byte is received at the same time, you can ignore the sent or received byte as you
' wish. For example, to do a read just send the dummy value $FF
' NOTE: assumes caller is controlling the chip select line
'
' INPUTS: spi_data8 - 8 bit data that is going to be sent
' OUTPUTS: returns the 8-bit data received at the same time
' -----------------------------------------------------------------------------

  ' reset result
  spi_result8 := 0

  ' shift 8-bits of data out while shifting in 8-bits of data
  ' data is sent and receiving MSB to LSB, so bit 7,6,5,4,3,2,1,0
  ' assume SPI interface is in mode 0/1, later make more flexible 
  repeat index from 0 to 7
    ' place next bit on data in pin to device
    OUTA [ spiDI ] := ((spi_data8 & $80) >> 7)
'   gfx_ntsc.Dec_Term((spi_data8 & $80) >> 7)
    
    spi_data8 <<= 1 ' shift data to left and prepare next bit for transmission

  ' pulse the clock line
    OUTA [ spiCLK ] := 1

      ' data is ready now on output line from device
    spi_result8 <<= 1            ' shift result to left
    spi_result8 |= INA [ spiDO ] ' read data bit and OR into result
                                    
    ' finish clock pulse
    OUTA [ spiCLK ] := 0

'  term.pstring(STRING("RECEIVED = "))
'  gfx_ntsc.Hex_Term((spi_result8, 2)
'  term.out($0D)

  ' return result
  return (spi_result8)

' end SPI_Send_Receive_Byte


' /////////////////////////////////////////////////////////////////////////////

PUB NES_Init_Gamepad

{{
Initializes the IO for the NES gamepad adapter

PARMS: None.
RETURNS: Nothing.


 DEFAULT SETUP (change the pin IO assignments if you move the inteface etc.)

     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0



' NES Bit Encoding
'
' RIGHT  = %00000001
' LEFT   = %00000010
' DOWN   = %00000100
' UP     = %00001000
' START  = %00010000
' SELECT = %00100000
' B      = %01000000
' A      = %10000000

}}

' step 1: set I/Os
DIRA [ NES_CLK ]   := 1 ' output
DIRA [ NES_LATCH ] := 1 ' output
DIRA [ NES_DATA1 ] := 0 ' input
DIRA [ NES_DATA0 ] := 0 ' input

' power IOs, we are using the PROP to feed power and GND to the GAMEPAD, it uses a 4021 CMOS chip, so only a few mA are needed,
' but if the controller has a lot of LEDs, you may need to manually connect 3.3V to the +5V
' note: the NES controller WILL work just fine with 3.3V which is what we use here

DIRA [ NES_5V ]  := 1 ' output
DIRA [ NES_GND ] := 1 ' output

OUTA [ NES_5V ]  := 1 ' set to 3.3V HIGH
OUTA [ NES_GND ] := 0 ' set to 0V LOW, virtual ground

' end NES_Init_Gamepad

' ////////////////////////////////////////////////////////////////////

PUB NES_Read_Gamepad : nes_bits |  i

{{
 Reads both gamepads in parallel encodes 8-bits for each in format
 right game pad #1 [15..8] : left game pad #0 [7..0]

 NOTE: Assumes IO pins have previously been setup (call to NES_Init_Gamepad eg.)

 PARMS: None.
 RETURNS: 16-bit NES button states in the format msb[8:NES gamepad 1 | 8:NES gamepad 0]lsb,
 where the 8-bits of each game controller is outlined below in the "NES bit encoding" table.


 DEFAULT SETUP (change the pin IO assignments if you move the inteface etc.)

                 |--------------------------------------|
                 |  Parallax NES Adapter                |
                 |                                      |
     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0

}}

' NES Bit Encoding
'
' RIGHT  = %00000001
' LEFT   = %00000010
' DOWN   = %00000100
' UP     = %00001000
' START  = %00010000
' SELECT = %00100000
' B      = %01000000
' A      = %10000000


' step 1: set clock and latch to 0
OUTA [ NES_CLK   ] := 0 
OUTA [ NES_LATCH ] := 0 
'Delay(1)

' step 2: set latch to 1
OUTA [ NES_LATCH ] := 1 
'Delay(1)

' step 3: set latch to 0
OUTA [ NES_LATCH ] := 0 

' step 4: read first bit of each game pad

' data is now ready to shift out
' first bit is ready 
nes_bits := 0

' left controller
nes_bits := INA[ NES_DATA0 ] | (INA[ NES_DATA1 ] << 8)

' step 5: read next 7 bits
repeat i from 0 to 6
 OUTA [ NES_CLK ] := 1
 'Delay(1)             
 OUTA [ NES_CLK ] := 0 
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[ NES_DATA0 ] | (INA[ NES_DATA1 ] << 8)

 'Delay(1)             
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' end NES_Read_Gamepad

' /////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY API (ported from my C/C++ libraries note syntax leftovers!)
' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Read_Status | _spi_data
{{
Reads the FLASH memory status register

PARMS: None.
RETURNS: status register value
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, READ_STATUS, $FF )

  ' now read data returned
  _spi_data := SPI_Write_Read( 8, $FF, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return results
  return ( _spi_data )   

' // end Flash_Read_Status

' /////////////////////////////////////////////////////////////////////////////


PUB Flash_Write_Status ( status ) | _spi_data
{{
Writes the FLASH memory status register

PARMS:

status - value to update the status register with

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_STATUS, $FF )
  _spi_data := SPI_Write_Read( 8, status, $FF )  
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return success for now
  return ( 1 )   

' // end Flash_Write_Status

' /////////////////////////////////////////////////////////////////////////////


PUB Flash_Open | _index, _spi_data, _flash_start_addr
{{

This functions "opens" the FLASH memory simply by unprotecting all sectors starting at address 0

PARMS: None.
RETURNS: None.

}}


  ' use global unprotect to unprotect entire chip at once, write "1" to bits 2,3,4,5
   'Flash_Write_Status ( Flash_Read_Status | %00_1111_00 )
   'Delay_MS( 10 )   

  
  ' this code is used to iterate thru all sectors manually and have more control
  ' iterate thru all sectors and unprotect them
  repeat _index from 0 to NUM_FLASH_ROM_SECTORS-1

    '// STEP 1: enable writing to chip  ///////////////////////////////////////
    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    Delay_MS( 10 )

    '// unprotect sector  ///////////////////////////////////

    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, UNPROTECT_SECTOR, $FF )

    '// get sector address
    _flash_start_addr := sector_addresses[ _index ]

    '// write sector address, there are only a handful of sectors since each is 4-64K
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    ' delay 10
    Delay_MS( 10 ) 

    ' // end for index

  ' // return success always
  return( 1 )

' // end Flash_Open

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Close | _index, _spi_data, _flash_start_addr      
{{
This function "closes" the FLASH by protecting all of the sectors on the flash chip.

PARMS: None.
RETURNS: None.

}}

  '// protecting the sectors seems to have a side effect, we are not considering???
  '// we need to read the data sheet more, for now, just return, and don't protect 
  '// all the sectors, this is overkill anyway for now ... 
  return(1)

  repeat _index from 0 to NUM_FLASH_ROM_SECTORS-1
      
    '// step 1: send WRITE_ENABLE command
        
    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
        
    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )
     
    Delay_MS(100)
     
    '// step 2: protect all sectors from sector address list
     
    '// enable SPI interface
    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
     
    _spi_data := SPI_Write_Read( 8, PROTECT_SECTOR, $FF )
     
    '// get sector address
    _flash_start_addr := sector_addresses[ _index ]

    '// write sector address, there are only a handful of sectors since each is 4-64K
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )
     
    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )
     
    Delay_MS(100)
     
    ' // end for index

  '// return success for now
  return(1)

' // end Flash_Close

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Erase_Chip( wait ) | _spi_data
{{
This function erases the entire chip. The erase command takes a considerable amount
of time, thus this function issues the command and either returns immediately or waits.
To determine if the erase is complete the function polls the status register bit 0
which is the RDY/BSY bit, 0=device is idle, 1=device is busy

PARMS: wait - boolean, 1=wait for erasure to complete, 0-return immediately.

RETURNS: None.

}}
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, CHIP_ERASE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' test if user requested wait for completion?
  if (wait)
    repeat
      ' erase operation still ongoing?
      if ((Flash_Read_Status & $01)==0)
        return (1)
     
  '// return success for now
  return(1)

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Erase( block ) | _index, _spi_data, _flash_start_addr
{{
This function erases a block of FLASH memory.

PARMS:

block - block address to erase.

RETURNS: None.

}}
  '// STEP 1: enable writing to chip again /////////////////////////////////////////////////////
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
  Delay_MS( 100 )
   
  '// STEP 2: erase 4K block of chip
  _flash_start_addr := FLASH_BLOCK_TO_ADDR ( block )
   
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, BLOCK_ERASE4, $FF )
   
  '// write block address, there are only a handful of blocks since each is 4-64K
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// erasure takes a long time... 
  Delay_MS( 500 )
   
  '// return success for now
  return(1)
   
' // end Flash_Erase

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function writes num_bytes bytes to the FLASH memory at flash_start_addr from the sent buffer_ptr.

PARMS:

flash_start_addr - starting address in the FLASH memory to write the data at.
buffer_ptr       - local byte pointer to Prop memory holding data to write to FLASH memory.
num_bytes        - number of bytes to write to FLASH from source buffer_ptr. 

}}

  '// STEP 1: enable writing to the chip again
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
  Delay_MS( 100 )
   
  '// STEP 2: write buffer to address in flash //////////////////
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
   
  '// initiate seq program command
  _spi_data := SPI_Write_Read( 8, SEQ_PROGRAM, $FF )
   
  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )
   
   
  '// write first byte
  _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ 0 ], $FF )
   
  '// de-assert set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// now write remaining data (if any)
  repeat _index from 1 to num_bytes-1
    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
     
    '// re-issue seq program command, but NO address needed
    _spi_data := SPI_Write_Read( 8,SEQ_PROGRAM,$FF )
     
    '// now data
    _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ _index ], $FF )
     
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

{
    ' debug info
    gfx_ntsc.Out_Term( "." )
}
     
    ' // end for index
     
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( SPI_CHAN_FLASH )
   
  '// finally send write disable command???
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_DISABLE, $FF )
   
  '// set CS to SPI select channel 0 (null)
   SPI_Select_Channel( 0 )
   
  '// we have written the data to the flash, now read it back and confirm its the same
  Delay_MS(1)
   
  '// return success
  return(1)
   
' // end Flash_Write

' ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Flash_Read( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function reads num_bytes from the FLASH memory starting at memory location flash_start_addr and stores the data in
buffer_ptr.

PARMS:

flash_start_addr - starting address in FLASH memory to read data from.
buffer_ptr       - local Prop byte pointer to store the read data at.
num_bytes        - the number of bytes to read from the FLASH memory. 

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )

  '// initiate read command
  _spi_data := SPI_Write_Read( 8, READ_DATA, $FF )

  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )
  
  '// now read data 
  repeat _index from  0 to num_bytes-1
    byte [buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' // end for index

  '// disable SPI interface

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// return success 
  return(1)

' // end Flash_Read

'//////////////////////////////////////////////////////////////////////////////
' FLASH MACROS ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB FLASH_ADDR_TO_BLOCK( n )
  n := (n >> 12)

' /////////////////////////////////////////////////////////////////////////////
  
PUB FLASH_BLOCK_TO_ADDR( n )
  n := (n << 12)

' /////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////
' SRAM API ////////////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Init ( mode ) | _spi_word
{{
This function initializes both banks of the Microchip 23K256 SRAMs
serial SPI SRAMs'. The initialization simply sets the STATUS register
of each SRAM into "byte", "page" or "sequential" read/write mode to allow more
efficient access of the SRAMs. The SRAM API currently relies on "sequential"
mode, so if yoo want to use them, this function should be called with SPI_MODE_SEQUENTIAL

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  mode - %00_00000_1 = SPI_MODE_BYTE
               %10_00000_1 = SPI_MODE_PAGE
               %01_00000_1 = SPI_MODE_SEQUENTIAL

RETURNS: Nothing. 

}}

  ' first set SRAM bank 0 to sequential access mode (see page 6 of data sheet)
  SPI_Select_Channel( SPI_CHAN_SRAM0 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' now select SRAM bank 1
  SPI_Select_Channel( SPI_CHAN_SRAM1 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' and finally de-select all SPI devices
  SPI_Select_Channel( 0 )

' end SRAM_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Write( bank, addr, num_bytes, src_buffer_ptr) | _index, _data, _spi_word
{{
This function writes a number of bytes to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        src_buffer_ptr - pointer to the buffer of bytes to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to write?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write sequential byte:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' and now send command to SRAM for sequential writing
    SPI_Write_Read( 24, _spi_word, $FF )
    
    ' write the bytes into the SRAM
    repeat _index from 0 to num_bytes-1

      ' get byte to write
      _data := byte[ src_buffer_ptr][ _index ]
       
      ' write the byte
      SPI_Write_Read( 8, _data, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Write

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Fill( bank, addr, num_bytes, value) | _index, _data, _spi_word
{{
This function writes a constant value to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        value          - byte value to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to fill?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write byte sequential:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' setup for sequential byte write, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to receive bytes (as many as we want to send)
    repeat _index from 0 to num_bytes-1
      SPI_Write_Read( 8, value, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Fill

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Read(bank, addr, num_bytes, dest_buffer_ptr) | _index, _data, _spi_word
{{

This function reads a number of bytes from either bank of the SRAM into a buffer.

The 23K256 supports a number of commands, we have only implemented a couple here
in these API functions (single byte read/write, multiple byte read/write).
However, there are many others you can implement, find out more by reviewing
the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  bank            - [0,1] the SRAM bank to read from, there are two banks, each holds 32K bytes
        addr            - address in SRAM to read from [0..32767]
        num_bytes       - number of bytes to read from SRAM into destination buffer
        dest_buffer_ptr - pointer to the buffer of bytes to read SRAM into

RETURNS: number of bytes read 

}}

  ' test for anything to read?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( read byte sequential:8 | address:16 )
    _spi_word := (%00000011 << 16) + (addr)

    ' setup for sequential byte read, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to be read from
    repeat _index from 0 to num_bytes-1

      ' read next byte from SRAM into receiving buffer
      byte [dest_buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )
    
    ' return bytes read
    return( num_bytes )
  else
    ' catch error
    return 0 

' end SRAM_Read

' /////////////////////////////////////////////////////////////////////////////                                                                               

PUB SRAM64K_MREAD( addr ) | _spi_word, _data
{{
This function reads a single byte from either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Read function
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
    
RETURNS: The 8-bit data from the SRAM. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( read byte sequential:8 | address:16 | dummy:8 )
    _spi_word := (%00000011 << 24) + (addr << 8) + 0 

    ' read the byte of data 
    _data := SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return data
    return ( _data )
    
' end SRAM64K_READ

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM64K_MWRITE( addr, data ) | _spi_word
{{
This function writes a single byte to either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Write|Fill functions
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
        data - the byte to write.    

RETURNS: Nothing. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( write sequential byte:8 | address:16 | data: 8)
    _spi_word := (%00000010 << 24) + ( (addr & $7FFF) << 8) + data

    ' write the byte of data 
    SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

' end SRAM64K_WRITE

' /////////////////////////////////////////////////////////////////////////////
' ANALOG TO DIGITAL API ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB AD_Init

{{
This function initializes the MCP3202 A/D converter. Currently, does nothing
more than asserts the CS then de-asserts it.

PARMS:  None.

RETURNS: Nothing. 

}}

  ' select and de-select the A/D device on the SPI bus just to wake it up and
  ' make sure we know the state of the A/D  
  SPI_Select_Channel( 0 )
  SPI_Select_Channel( SPI_CHAN_AD )
  SPI_Select_Channel( 0 )

' end AD_Init

' /////////////////////////////////////////////////////////////////////////////

PUB AD_Read( channel ) | _data

{{
This function reads from the MCP3202 the sent AD channel (0,1) in single ended mode and returns a
12-bit integer [0..4095] representing the value.

PARMS:  channel - [0,1] which channel to read. 

RETURNS: returns value of A/D converter [0..4095] 

}}

  ' read channel 0, SPI transaction is 17 bits total, control bits indicate single ended mode, channel (0,1)
  SPI_Select_Channel( SPI_CHAN_AD )
  _data := SPI_Write_Read( 17, %0001_10_1_0000000000000 | channel << 14, $FFF )
  SPI_Select_Channel( 0 )

  ' return A/D value
  return ( _data )
   
' end AD_Read

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////


PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed
sram_test_vector        byte      $FF, $00, $55, $AA ' test vector of bytes used to test SRAM
{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}
